require 'net/http'
require 'json'

module KeyholeIO
  APIURL = "http://api.keyhole.io/v1"


  class Response
    attr_accessor :meta, :data
    
    def initialize response
      resp = JSON.parse response
      @meta = resp["meta"]
      if @meta["status"] >= 200 and @meta["status"] <= 299
        @data = resp["data"]
      end
    end

    def successful?
      @meta["status"] >= 200 and @meta["status"] <= 299
    end   
  end


  class KVPair
    attr_accessor :meta, :key, :value, :meta, :created_at, :updated_at, :type, :successful

    def initialize raw_response
      response = Response.new raw_response
      if response.successful?
        @meta = response.meta
        @key = response.data["key"]
        @value = response.data["value"]
        @created_at = response.data["created_at"] if response.data["created_at"]
        @updated_at = response.data["updated_at"] if response.data["updated_at"]
        @type = response.data["type"] if response.data["type"]
        @successful = true
      else
        @successful = false
      end
    end

    def successful?
      @successful
    end
  end


  class Group
    attr_accessor :id, :apikey

    def initialize name, apikey
      @apikey = apikey
      @id = name

      response = Net::HTTP.post_form URI.parse("#{APIURL}/group"), paramshash(name)
      r = Response.new response.body
    end

    def entities
      response = Net::HTTP.get URI.parse("#{APIURL}/group?apikey=#{@apikey}&id=#{@id}")
      Response.new response
    end

    def add entity
      if entity.class == Entity
        http = Net::HTTP.new("api.keyhole.io")
        req = Net::HTTP::Put.new("/v1/group")
        req.set_form_data apikey: @apikey, id: @id, entity: entity.name

        Response.new http.request(req).body
      else
        puts "'entity' needs to be of type Entity."
      end
    end

    def remove entity
      if entity.class == Entity
        http = Net::HTTP.new("api.keyhole.io")
        req = Net::HTTP::Delete.new("/v1/group?apikey=#{@apikey}&id=#{@id}&entity=#{entity.name}")
        Response.new http.request(req).body
      else
        puts "'entity' needs to be of type Entity."        
      end
    end

    def destroy
      http = Net::HTTP.new("api.keyhole.io")
      req = Net::HTTP::Delete.new("/v1/group/rm?apikey=#{@apikey}&id=#{@id}")
      Response.new http.request(req).body
    end

    private

    def paramshash name, options = {}
      self.class.paramshash name, @apikey, options
    end

    def self.paramshash name, apikey, options
      params = { 
        apikey: apikey,
        id: name
      }
      options.collect { |k,v| params[k] = v }

      params
    end
  end

  
  class Entity
    attr_accessor :token, :id, :apikey, :name, :channel_name

    ##
    ## Entity Methods
    ##

    def initialize name, apikey, options = {}
      # options: desc, password, public      
      @apikey = apikey
      @name = name
      
      response = Net::HTTP.post_form URI.parse("#{APIURL}/entity"), paramshash(name, options)
      r = Response.new response.body
      if r.successful?
        @token = r.data["token"]
      end
    end

    def info options = {}
      self.class.get_info @name, @apikey, options
    end

    def destroy
      http = Net::HTTP.new("api.keyhole.io")
      req = Net::HTTP::Delete.new("/v1/entity?token=#{@token}&apikey=#{@apikey}&id=#{@name}")
      Response.new http.request(req).body
    end

    def nullify_token
      http = Net::HTTP.new("api.keyhole.io")
      req = Net::HTTP::Put.new("/v1/entity")
      req.set_form_data token: @token

      Response.new http.request(req).body
    end

    def set_password new_pass
      response = Net::HTTP.post_form URI.parse("#{APIURL}/entity/password"), tokenparams(new: new_pass)
      Response.new response.body
    end

    def change_password old_pass, new_pass
      response = Net::HTTP.post_form URI.parse("#{APIURL}/entity/password"), tokenparams(old: old_pass, new: new_pass)
      Response.new response.body
    end

    def delete key
      http = Net::HTTP.new("api.keyhole.io")
      req = Net::HTTP::Delete.new("/v1/datastore/#{key}?token=#{@token}")
      Response.new http.request(req).body
    end

    ##
    ## Channel Methods
    ##

    def channel
      response = Net::HTTP.post_form URI.parse("#{APIURL}/entity/channel"), tokenparams(nil)
      r = Response.new response.body
      if r.successful?
        @channel_name = r.data["name"]
      end
    end

    def channel_connected
      response = Net::HTTP.get URI.parse("#{APIURL}/entity/channel?token=#{@token}")
      Response.new response
    end

    def channel_rename
      http = Net::HTTP.new("api.keyhole.io")
      req = Net::HTTP::Put.new("/v1/entity/channel")
      req.set_form_data token: @token

      r = Response.new http.request(req).body
      if r.successful?
        @channel_name = r.data["name"]
      end
    end

    def channel_destroy
      http = Net::HTTP.new("api.keyhole.io")
      req = Net::HTTP::Delete.new("/v1/entity/channel?token=#{@token}")
      r = Response.new http.request(req).body
      if r.successful?
        @channel_name = nil
      end
      r
    end


    ##
    ## Messaging Client Methods
    ##

    def add_client id, type
      response = Net::HTTP.post_form URI.parse("#{APIURL}/entity/messaging/#{type}"), tokenparams(id: id)
      Response.new response.body
    end

    def add_email_client id
      add_client id, "email"
    end

    def add_apns_client id
      add_client id, "apns"
    end

    def check_client id, type
      response = Net::HTTP.get URI.parse("#{APIURL}/entity/messaging/#{type}?token=#{@token}&id=#{id}")
      r = Response.new response
      if r.successful?
        r.meta["status"] == 200
      else
        false
      end
    end

    def email_client_exists? id
      check_client id, "email"
    end

    def apns_client_exists? id
      check_client id, "apns"
    end

    def delete_client id, type
      http = Net::HTTP.new("api.keyhole.io")
      req = Net::HTTP::Delete.new("/v1/entity/messaging/#{type}?token=#{@token}&id=#{id}")
      r = Response.new http.request(req).body
    end

    def remove_email_client id
      delete_client id, "email"
    end

    def remove_apns_client id
      delete_client id, "apns"
    end

    ## 
    ## Datastore Methods
    ## 

    def set kvpair
      tparams = {}
      unless kvpair.empty?
        tparams[:k] = kvpair.keys.first
        if (kvpair.values.first.class == Array or kvpair.values.first.class == Hash)
          tparams[:v] = kvpair.values.first.to_json
        else  
          tparams[:v] = kvpair.values.first
        end
      end

      if kvpair.size > 1
        tparams[:meta] = kvpair.values.last.to_json
      end

      response = Net::HTTP.post_form URI.parse("#{APIURL}/datastore"), tokenparams(tparams)
      Response.new response.body
    end

    def datastores options = {}
      response = Net::HTTP.get URI.parse("#{APIURL}/datastore?#{tokenparams_string(options)}")
      Response.new response
    end

    def get key
      response = Net::HTTP.get URI.parse("#{APIURL}/datastore/#{key}?#{tokenparams_string(nil)}")
      KVPair.new response
    end

    def remove key
      http = Net::HTTP.new("api.keyhole.io")
      req = Net::HTTP::Delete.new("/v1/datastore/#{key}?token=#{@token}")
      r = Response.new http.request(req).body
    end

    ##
    ## Class Methods
    ##

    def self.get_info name, apikey, options = {}
      # options: password
      response = Net::HTTP.get(URI.parse("#{APIURL}/entity?#{self.paramshash_string(name, apikey, options)}"))
      Response.new response
    end

    #private 

    def tokenparams options = {}
      params = { token: @token }
      options.collect { |k,v| params[k] = v } unless options.nil?
      params      
    end

    def tokenparams_string options
      tokenparams(options).collect { |k, v| "#{k}=#{v}" }.join("&")
    end

    def paramshash name, options = {}
      self.class.paramshash name, @apikey, options
    end

    def self.paramshash name, apikey, options
      params = { 
        apikey: apikey,
        id: name
      }
      options.collect { |k,v| params[k] = v }

      params
    end

    def paramshash_string name, options = {}
      self.class.paramshash_string name, options, @apikey
    end

    def self.paramshash_string name, apikey, options
      self.paramshash(name, apikey, options).collect { |k, v| "#{k}=#{v}" }.join("&")
    end

  end

end